from typing import Any, Callable, Optional, TypeVar, Union, cast, overload, List, Dict

import ipyvue
import ipyvuetify as vw
import reacton
from typing_extensions import Literal

import solara
from solara.alias import rv as v

T = TypeVar("T")


def use_change(el: reacton.core.Element, on_value: Callable[[Any], Any], enabled=True, update_events=["blur", "keyup.enter"]):
    """Trigger a callback when a blur events occurs or the enter key is pressed."""
    on_value_ref = solara.use_ref(on_value)
    on_value_ref.current = on_value

    def add_events():
        def on_change(widget, event, data):
            if enabled:
                on_value_ref.current(widget.v_model)

        widget = cast(ipyvue.VueWidget, solara.get_widget(el))
        if enabled:
            for event in update_events:
                widget.on_event(event, on_change)

        def cleanup():
            if enabled:
                for event in update_events:
                    widget.on_event(event, on_change, remove=True)

        return cleanup

    solara.use_effect(add_events, [enabled])


@solara.component
def InputText(
    label: str,
    value: Union[str, solara.Reactive[str]] = "",
    on_value: Callable[[str], None] = None,
    disabled: bool = False,
    password: bool = False,
    continuous_update: bool = False,
    update_events: List[str] = ["blur", "keyup.enter"],
    error: Union[bool, str] = False,
    message: Optional[str] = None,
    classes: List[str] = [],
    style: Optional[Union[str, Dict[str, str]]] = None,
):
    """Free form text input.

    ### Basic example:

    ```solara
    import solara

    text = solara.reactive("Hello world!")
    continuous_update = solara.reactive(True)

    @solara.component
    def Page():
        solara.Checkbox(label="Continuous update", value=continuous_update)
        solara.InputText("Enter some text", value=text, continuous_update=continuous_update.value)
        with solara.Row():
            solara.Button("Clear", on_click=lambda: text.set(""))
            solara.Button("Reset", on_click=lambda: text.set("Hello world"))
        solara.Markdown(f"**You entered**: {text.value}")
    ```

    ### Password input:

    This will not show the entered text.

    ```solara
    import solara

    password = solara.reactive("Super secret")
    continuous_update = solara.reactive(True)

    @solara.component
    def Page():
        solara.Checkbox(label="Continuous update", value=continuous_update)
        solara.InputText("Enter a passsword", value=password, continuous_update=continuous_update.value, password=True)
        with solara.Row():
            solara.Button("Clear", on_click=lambda: password.set(""))
            solara.Button("Reset", on_click=lambda: password.set("Super secret"))
        solara.Markdown(f"**You entered**: {password.value}")
    ```


    ## Arguments

    * `label`: Label to display next to the slider.
    * `value`: The currently entered value.
    * `on_value`: Callback to call when the value changes.
    * `disabled`: Whether the input is disabled.
    * `password`: Whether the input is a password input (typically shows input text obscured with an asterisk).
    * `continuous_update`: Whether to call the `on_value` callback on every change or only when the input loses focus or the enter key is pressed.
    * `update_events`: A list of events that should trigger `on_value`. If continuous update is enabled, this will effectively be ignored,
        since updates will happen every change.
    * `error`: If truthy, show the input as having an error (in red). If a string is passed, it will be shown as the error message.
    * `message`: Message to show below the input. If `error` is a string, this will be ignored.
    * `classes`: List of CSS classes to apply to the input.
    * `style`: CSS style to apply to the input.
    """
    reactive_value = solara.use_reactive(value, on_value)
    del value, on_value
    style_flat = solara.util._flatten_style(style)
    classes_flat = solara.util._combine_classes(classes)

    def set_value_cast(value):
        reactive_value.value = str(value)

    def on_v_model(value):
        if continuous_update:
            set_value_cast(value)

    messages = []
    if error and isinstance(error, str):
        messages.append(error)
    elif message:
        messages.append(message)
    text_field = v.TextField(
        v_model=reactive_value.value,
        on_v_model=on_v_model,
        label=label,
        disabled=disabled,
        type="password" if password else None,
        error=bool(error),
        messages=messages,
        class_=classes_flat,
        style_=style_flat,
    )
    use_change(text_field, set_value_cast, enabled=not continuous_update, update_events=update_events)
    return text_field


@overload
@solara.component
def InputFloat(
    label: str,
    value: Union[float, solara.Reactive[float]] = 0,
    on_value: Optional[Callable[[float], None]] = ...,
    disabled: bool = ...,
    optional: Literal[False] = ...,
    continuous_update: bool = ...,
    clearable: bool = ...,
    classes: List[str] = ...,
    style: Optional[Union[str, Dict[str, str]]] = ...,
) -> reacton.core.ValueElement[vw.TextField, Any]:
    ...


@overload
@solara.component
def InputFloat(
    label: str,
    value: Union[Optional[float], solara.Reactive[Optional[float]]] = 0,
    on_value: Optional[Callable[[Optional[float]], None]] = ...,
    disabled: bool = ...,
    optional: Literal[True] = ...,
    continuous_update: bool = ...,
    clearable: bool = ...,
    classes: List[str] = ...,
    style: Optional[Union[str, Dict[str, str]]] = ...,
) -> reacton.core.ValueElement[vw.TextField, Any]:
    ...


@solara.component
def InputFloat(
    label: str,
    value: Union[None, float, solara.Reactive[float], solara.Reactive[Optional[float]]] = 0,
    on_value: Union[None, Callable[[Optional[float]], None], Callable[[float], None]] = None,
    disabled: bool = False,
    optional: bool = False,
    continuous_update: bool = False,
    clearable: bool = False,
    classes: List[str] = [],
    style: Optional[Union[str, Dict[str, str]]] = None,
):
    """Numeric input (floats).

    Basic example:

    ```solara
    import solara

    float_value = solara.reactive(42.0)
    continuous_update = solara.reactive(True)

    @solara.component
    def Page():
        solara.Checkbox(label="Continuous update", value=continuous_update)
        solara.InputFloat("Enter a float number", value=float_value, continuous_update=continuous_update.value)
        with solara.Row():
            solara.Button("Clear", on_click=lambda: float_value.set(42.0))
        solara.Markdown(f"**You entered**: {float_value.value}")
    ```


    ## Arguments

    * `label`: Label to display next to the slider.
    * `value`: The currently entered value.
    * `on_value`: Callback to call when the value changes.
    * `disabled`: Whether the input is disabled.
    * `optional`: Whether the value can be None.
    * `continuous_update`: Whether to call the `on_value` callback on every change or only when the input loses focus or the enter key is pressed.
    * `clearable`: Whether the input can be cleared.
    * `classes`: List of CSS classes to apply to the input.
    * `style`: CSS style to apply to the input.

    """

    def str_to_float(value: Optional[str]):
        if value:
            try:
                value = value.replace(",", ".")
                return float(value)
            except ValueError:
                raise ValueError("Value must be a number")
        else:
            if optional:
                return None
            else:
                raise ValueError("Value cannot be empty")

    return _InputNumeric(
        str_to_float,
        label=label,
        value=value,
        on_value=on_value,
        disabled=disabled,
        continuous_update=continuous_update,
        clearable=clearable,
        classes=classes,
        style=style,
    )


@overload
@solara.component
def InputInt(
    label: str,
    value: Union[int, solara.Reactive[int]] = 0,
    on_value: Optional[Callable[[int], None]] = ...,
    disabled: bool = ...,
    optional: Literal[False] = ...,
    continuous_update: bool = ...,
    clearable: bool = ...,
    classes: List[str] = ...,
    style: Optional[Union[str, Dict[str, str]]] = ...,
) -> reacton.core.ValueElement[vw.TextField, Any]:
    ...


@overload
@solara.component
def InputInt(
    label: str,
    value: Union[Optional[int], solara.Reactive[Optional[int]]] = 0,
    on_value: Optional[Callable[[Optional[int]], None]] = ...,
    disabled: bool = ...,
    optional: Literal[True] = ...,
    continuous_update: bool = ...,
    clearable: bool = ...,
    classes: List[str] = ...,
    style: Optional[Union[str, Dict[str, str]]] = ...,
) -> reacton.core.ValueElement[vw.TextField, Any]:
    ...


@solara.component
def InputInt(
    label: str,
    value: Union[None, int, solara.Reactive[int], solara.Reactive[Optional[int]]] = 0,
    on_value: Union[None, Callable[[Optional[int]], None], Callable[[int], None]] = None,
    disabled: bool = False,
    optional: bool = False,
    continuous_update: bool = False,
    clearable: bool = False,
    classes: List[str] = [],
    style: Optional[Union[str, Dict[str, str]]] = None,
):
    """Numeric input (integers).

    Basic example:

    ```solara
    import solara

    int_value = solara.reactive(42)
    continuous_update = solara.reactive(True)

    @solara.component
    def Page():
        solara.Checkbox(label="Continuous update", value=continuous_update)
        solara.InputInt("Enter an integer number", value=int_value, continuous_update=continuous_update.value)
        with solara.Row():
            solara.Button("Clear", on_click=lambda: int_value.set(42))
        solara.Markdown(f"**You entered**: {int_value.value}")
    ```

    ## Arguments

    * `label`: Label to display next to the slider.
    * `value`: The currently entered value.
    * `on_value`: Callback to call when the value changes.
    * `disabled`: Whether the input is disabled.
    * `optional`: Whether the value can be None.
    * `continuous_update`: Whether to call the `on_value` callback on every change or only when the input loses focus or the enter key is pressed.
    * `clearable`: Whether the input can be cleared.
    * `classes`: List of CSS classes to apply to the input.
    * `style`: CSS style to apply to the input.
    """

    def str_to_int(value: Optional[str]):
        if value:
            try:
                return int(value)
            except ValueError:
                raise ValueError("Value must be an integer")
        else:
            if optional:
                return None
            else:
                raise ValueError("Value cannot be empty")

    return _InputNumeric(
        str_to_int,
        label=label,
        value=value,
        on_value=on_value,
        disabled=disabled,
        continuous_update=continuous_update,
        clearable=clearable,
        classes=classes,
        style=style,
    )


def _use_input_type(
    input_value: Union[None, T, solara.Reactive[Optional[T]], solara.Reactive[T]],
    parse: Callable[[Optional[str]], T],
    stringify: Callable[[Optional[T]], str],
    on_value: Union[None, Callable[[Optional[T]], None], Callable[[T], None]] = None,
):
    reactive_value = solara.use_reactive(input_value, on_value)  # type: ignore
    del input_value, on_value
    string_value, set_string_value = solara.use_state(stringify(reactive_value.value) if reactive_value.value is not None else None)
    # Use a ref to make sure sync_back_input_value() does not get a stale string_value
    string_value_ref = solara.use_ref(string_value)
    string_value_ref.current = string_value

    error_message = cast(Union[str, None], None)

    try:
        reactive_value.set(parse(string_value))
    except ValueError as e:
        error_message = str(e.args[0])

    def sync_back_input_value():
        def on_external_value_change(new_value: Optional[T]):
            new_string_value = stringify(new_value)
            try:
                parse(string_value_ref.current)
            except ValueError:
                # String value could be invalid when external value is changed by a different component
                set_string_value(new_string_value)
            else:
                if new_value != parse(string_value_ref.current):
                    set_string_value(new_string_value)

        return reactive_value.subscribe(on_external_value_change)

    solara.use_effect(sync_back_input_value, [reactive_value])

    return string_value, error_message, set_string_value


@solara.component
def _InputNumeric(
    str_to_numeric: Callable[[Optional[str]], T],
    label: str,
    value: Union[None, T, solara.Reactive[Optional[T]], solara.Reactive[T]],
    on_value: Union[None, Callable[[Optional[T]], None], Callable[[T], None]] = None,
    disabled: bool = False,
    continuous_update: bool = False,
    clearable: bool = False,
    classes: List[str] = [],
    style: Optional[Union[str, Dict[str, str]]] = None,
):
    """Numeric input.

    ## Arguments

    * `label`: Label to display next to the slider.
    * `value`: The currently entered value.
    * `on_value`: Callback to call when the value changes.
    * `disabled`: Whether the input is disabled.
    * `continuous_update`: Whether to call the `on_value` callback on every change or only when the input loses focus or the enter key is pressed.
    * `classes`: List of CSS classes to apply to the input.
    * `style`: CSS style to apply to the input.
    """
    style_flat = solara.util._flatten_style(style)
    classes_flat = solara.util._combine_classes(classes)

    internal_value, error, set_value_cast = _use_input_type(
        value,
        str_to_numeric,
        str,
        on_value,
    )

    def on_v_model(value):
        if continuous_update:
            set_value_cast(value)

    if error:
        label += f" ({error})"
    text_field = v.TextField(
        v_model=internal_value,
        on_v_model=on_v_model,
        label=label,
        disabled=disabled,
        # we are not using the number type, since we cannot validate invalid input
        # see https://stackoverflow.blog/2022/12/26/why-the-number-input-is-the-worst-input/
        # type="number",
        hide_details=True,
        clearable=clearable,
        error=bool(error),
        class_=classes_flat,
        style_=style_flat,
    )
    use_change(text_field, set_value_cast, enabled=not continuous_update)
    return text_field



"""
# Themes in Solara

Theming is provided to Solara through the [`ipyvuetify`](https://solara.dev/docs/understanding/ipyvuetify) package.
Two themes are provided by default: light and dark.
Control over the theme variant (light, dark, or auto) can be provided to the user through the `ThemeToggle` component. Developers can
control theming through the `solara.lab.theme` object.

## Themes

The default themes can be customized through altering `solara.lab.theme`. The
[theme options of vuetify](https://v2.vuetifyjs.com/en/features/theme/#customizing)
are available as `solara.lab.theme.themes.light` and `solara.lab.theme.themes.dark`. The different properties of the theme can be set through, for example

```python
solara.lab.theme.themes.light.primary = "#3f51b5"
```

Dark theme can be enabled/disabled through `solara.lab.theme.dark = True` / `False` / `None`. When set to `None`, the theme is set to auto,
which will follow the system theme if the user's browser supports detecting it.

When using solara server, the default theme variant can be set in the server settings through the command line interface. See
[here](https://solara.dev/docs/understanding/solara-server) for more information.

The active theme can be detected through `solara.lab.theme.dark_effective`, which is set to `True` or `False` depending on
whether dark theme is enabled or not.

## Example

```solara
import solara
import solara.lab
from solara.lab import theme as theme


def change_color(colors):
    if "purple" in colors:
        theme.themes.light.info = "#8617c2"
    else:
        theme.themes.light.info = "#2196f3"

    if "green" in colors:
        theme.themes.light.error = "#33bd65"
    else:
        theme.themes.light.error = "#ff5252"


@solara.component
def Page():
    solara.Info("Info message")
    solara.Error("Error message")

    with solara.ToggleButtonsMultiple(on_value=change_color):
        solara.Button("Change Info", value="purple")
        solara.Button("Change Error", value="green")
```

A more advanced example can be found at [the scatter app demo](https://solara.dev/apps/scatter) where
we set a plotly theme based on dark/light theme, and also set the app bar color to its default so it
responds to the theme variant change.

## ThemeToggle

"""

import solara
from solara.website.utils import apidoc

from . import NoPage

title = "Themes"
Page = NoPage
__doc__ += apidoc(solara.lab.ThemeToggle)  # type: ignore



import plotly.express as px

import solara
import solara.express

df = px.data.iris()


def test_cross_filter():
    filter, set_filter = None, None

    @solara.component
    def Test():
        nonlocal filter, set_filter
        solara.provide_cross_filter()
        filter, set_filter = solara.use_cross_filter(id(df))

        with solara.HBox() as main:
            solara.express.scatter(df, x="sepal_length", y="sepal_width")
            solara.express.scatter(df, x="sepal_length", y="sepal_width", size=[10 for ea in df.sepal_length], log_x=True)
        return main

    box, rc = solara.render(Test(), handle_error=False)
    assert set_filter is not None
    set_filter(df["sepal_length"] > 5)



"""

# Component.widget

Create a classic ipywidget from a component.

```python
def widget(self, **kwargs):
    ...
```


This will create a widget from the component. The widget will be a
subclass of `ipywidgets.VBox` and `ipywidgets.ValueWidget`.

Example

```python
import solara
widget = solara.FileDownload.widget(data="some text data", filename="solara-demo.txt")
```

This is very useful if you are migrating your application from a classic
ipywidget to solara. See [also the ipywidgets tutorial](/docs/tutorial/ipywidgets).

The `ipywidgets.ValueWidget` is used to enable the use of the widget in
interact, or interactive. The `ipywidgets.VBox` is used to enable
nesting of widgets.

All keyword arguments will be passed to the component.
Each argument pair of `on_<name>` and `<name>` will
be added as a trait in the widget.

For example,

```solara
import solara
import ipywidgets as widgets
import random


countries_demo_data = {
    "Netherlands": ["Amsterdam", "Rotterdam", "The Hague"],
    "Germany": ["Berlin", "Hamburg", "Munich"],
    "France": ["Paris", "Marseille", "Lyon"],
}


# this component can be used in a component three, but ...
@solara.component
def LocationSelect(value, on_value, countries=countries_demo_data):
    country, city = value
    cities = countries.get(country, [])
    # reset to None if not in the list of countries
    if city not in cities:
        city = None
        # update the state if we changed/reset city
        on_value((country, city))

    with solara.Card("Location"):
        solara.Select(label="country",
                      values=list(countries),
                      value=country,
                      on_value=lambda country: on_value((country, city)),
        )
        solara.Select(label="city",
                      values=cities,
                      value=city,
                      on_value=lambda city: on_value((country, city)),
        )


# Using .widget(...) we can create a widget from it.
# For use with interact:
@widgets.interact(location=LocationSelect.widget(value=("Netherlands", "Amsterdam")))
def f(size=3.4, location=None):
    print(size, location)


# Or to add to your VBox:
widgets.VBox(
    [LocationSelect.widget(value=("Netherlands", "Amsterdam"))]
)

# this is how you'd use it as a component
@solara.component
def Page():
    value, set_value = solara.use_state(("Netherlands", "Amsterdam"))
    def pick():
        country = random.choice(list(countries_demo_data))
        city = random.choice(countries_demo_data[country])
        set_value((country, city))

    LocationSelect(value=value, on_value=set_value)
    solara.Button("Pick random place", on_click=pick)

```

"""

from . import NoPage

Page = NoPage
title = "widget"



"""
# PivotTable

Interactive [Pivot Table](https://en.wikipedia.org/wiki/Pivot_table) showing aggregated data in groups.

We provide three version of the pivot table

    * PivotTableView: Takes already aggregated data as input
    * PivotTable: Takes a dataframe as input and aggregates it for you.
    * PivotTableCard: Similar to PivotTable but with a card layout, and configuration options.

"""

import solara

try:
    import vaex.datasets

    df = vaex.datasets.titanic()
except Exception:
    df = None
    vaex = None


@solara.component
def View():
    # select on the 2nd x axis, index=3, which means (pclass=2, survived=true)
    # Note: the indices refer to the header_x and header_y
    selected, on_selected = solara.use_state({"x": [1, 3]})
    data = solara.PivotTableData(
        {
            "x": ["pclass", "survived"],
            "y": ["sex"],
            "agg": "count",
            "values": [["5", "118"], ["139", "61"], ["12", "146"], ["94", "25"], ["110", "418"], ["106", "75"]],
            "values_x": ["123", "200", "158", "119", "528", "181"],
            "values_y": ["466", "843"],
            "headers_x": [["1", "1", "2", "2", "3", "3"], ["false", "true", "false", "true", "false", "true"]],
            "headers_y": [["female", "male"]],
            "counts_x": 6,
            "counts_y": 2,
            "total": "1,309",
        }
    )
    with solara.VBox() as main:
        solara.Markdown(f"`selected = {selected}`")
        solara.PivotTableView(data=data, selected=selected, on_selected=on_selected)
    return main


@solara.component
def Page():
    if vaex is None:
        return solara.Markdown("This example requires vaex, please install it with `pip install vaex`")
    with solara.Div() as main:
        solara.Markdown("# Titanic")
        selected, on_selected = solara.use_state({"x": [0, 0]})
        solara.provide_cross_filter()
        with solara.VBox():
            type, set_type = solara.use_state("view")
            with solara.ToggleButtonsSingle(type, on_value=set_type):
                solara.Button("PivotTableView", value="view")
                solara.Button("PivotTable", value="df")
                solara.Button("PivotTableCard", value="card")
            if type == "view":
                solara.Markdown("# PivotTableView\nThis component will take aggregates data as input")
                View()
            elif type == "df":
                solara.Markdown("# PivotTable\nThis component aggregates the dataframe for you")
                solara.Markdown(f"`selected = {selected}`")
                solara.PivotTable(df, ["pclass"], ["sex"], selected=selected, on_selected=on_selected)
            elif type == "card":
                solara.Markdown("# PivotTable\nThis component aggregates the dataframe for you, and gives a UI to configure the component")
                solara.Markdown(f"`selected = {selected}`")
                solara.PivotTableCard(df, ["pclass"], ["sex"], selected=selected, on_selected=on_selected)
    return main


Component = Page
App = Component
app = Page()



"""
# on_kernel_start

Run a function when a virtual kernel (re)starts and optionally run a cleanup function on shutdown.

```python
def on_kernel_start(f: Callable[[], Optional[Callable[[], None]]]):
```

`f` will be called on each virtual kernel (re)start. This (usually) happens each time a browser tab connects to the server
[see solara server for more details](https://solara.dev/docs/understanding/solara-server).
The (optional) function returned by `f` will be called on kernel shutdown.

Note that the cleanup functions are called in reverse order with respect to the order in which they were registered
(e.g. the cleanup function of the last call to `on_kernel_start` will be called first on kernel shutdown)
"""

from . import NoPage

title = "on_kernel_start"
Page = NoPage



import reacton  # noqa: F401
from reacton import ipyvue as vue  # noqa: F401
from reacton import ipyvuetify as v  # noqa: F401
from reacton import ipywidgets as w  # noqa: F401

import solara as sol  # noqa: F401

from . import layout  # noqa: F401



"""
# ColumnsResponsive

"""
import solara
import solara.lab
from solara.website.utils import apidoc

title = "ColumnsResponsive"

gutters = solara.reactive(True)
gutters_dense = solara.reactive(True)
wrap = solara.reactive(True)

children_count = solara.reactive(12)
columns_default = solara.reactive(6)
columns_small = solara.reactive(4)
columns_medium = solara.reactive(3)
columns_large = solara.reactive(2)
columns_xlarge = solara.reactive(1)


@solara.component
def Page():
    with solara.VBox() as main:
        with solara.Card("Controls"):
            solara.Checkbox(label="Wrap").connect(wrap)
            solara.Checkbox(label="Gutters").connect(gutters)
            solara.Checkbox(label="Dense gutters").connect(gutters_dense)
            solara.IntSlider("Children", max=20).connect(children_count)

            solara.Select("columns default", values=[1, 2, 3, 4, 6, 12]).connect(columns_default)  # type: ignore
            solara.Select("columns small", values=[1, 2, 3, 4, 6, 12]).connect(columns_small)  # type: ignore
            solara.Select("columns medium", values=[1, 2, 3, 4, 6, 12]).connect(columns_medium)  # type: ignore
            solara.Select("columns large", values=[1, 2, 3, 4, 6, 12]).connect(columns_large)  # type: ignore
            solara.Select("columns xlarge", values=[1, 2, 3, 4, 6, 12]).connect(columns_xlarge)  # type: ignore
        # taken from https://v2.vuetifyjs.com/en/styles/display/#display
        solara.HTML(
            "h2", unsafe_innerHTML=f"Current screensize is xsmall/default, each child is {columns_default.value} points wide", class_="ma-2 d-flex d-sm-none ma"
        )
        solara.HTML(
            "h2", unsafe_innerHTML=f"Current screensize is small, each child is {columns_small.value} points wide", class_="ma-2 d-none d-sm-flex d-md-none"
        )
        solara.HTML(
            "h2", unsafe_innerHTML=f"Current screensize is medium, each child is {columns_medium.value} points wide", class_="ma-2 d-none d-md-flex d-lg-none"
        )
        solara.HTML(
            "h2", unsafe_innerHTML=f"Current screensize is large, each child is {columns_large.value} points wide", class_="ma-2 d-none d-lg-flex d-xl-none"
        )
        solara.HTML("h2", unsafe_innerHTML=f"Current screensize is xlarge, each child is {columns_xlarge.value} points wide", class_="ma-2 d-none d-xl-flex")
        solara.Markdown("Change the screen size to see the effect of the different columns sizes.")
        with solara.ColumnsResponsive(
            default=columns_default.value,
            small=columns_small.value,
            medium=columns_medium.value,
            large=columns_large.value,
            xlarge=columns_xlarge.value,
            wrap=wrap.value,
            gutters=gutters.value,
            gutters_dense=gutters_dense.value,
        ):
            for i in range(children_count.value):
                solara.Text(f"{i}")
    return main


__doc__ += apidoc(solara.ColumnsResponsive.f)  # type: ignore



"""
# Columns

"""
import solara
import solara.lab
from solara.website.utils import apidoc

gutters = solara.reactive(True)
gutters_dense = solara.reactive(True)


@solara.component
def Page():
    with solara.Columns([1, 2, 1], gutters=gutters.value, gutters_dense=gutters_dense.value) as main:
        with solara.Card("Left", margin=0):
            solara.Checkbox(label="Gutters").connect(gutters)
            solara.Checkbox(label="Dense gutters").connect(gutters_dense)
        with solara.Card("Middle", margin=0):
            solara.Markdown("This column has a relative width of 2, the columns to the left and right have a relative width of 1.")
        with solara.Card("Right", margin=0):
            solara.Markdown("This column has a relative width of 1, the columns to the left has a relative width of 2.")
    return main


__doc__ += apidoc(solara.Columns.f)  # type: ignore



from typing import Optional

import solara
import vaex

df = vaex.from_arrays(x=[0, 1, 2, 3, 4], y=[0, 1, 4, 9, 16])


def test_basic():
    set_filter1 = filter1 = None
    set_filter2 = filter2 = None

    @solara.component
    def FilterDummy1():
        nonlocal filter1, set_filter1
        filter1, set_filter1 = solara.use_cross_filter(id(df), "test1")
        return solara.Text("dummy1")

    @solara.component
    def FilterDummy2():
        nonlocal filter2, set_filter2
        filter2, set_filter2 = solara.use_cross_filter(id(df), "test1")
        return solara.Text("dummy2")

    @solara.component
    def Page():
        solara.provide_cross_filter()
        with solara.VBox() as main:
            FilterDummy1()
            FilterDummy2()
        return main

    rc, box = solara.render(Page(), handle_error=False)
    assert set_filter1 is not None
    assert set_filter2 is not None
    set_filter1(df.x > 1)
    assert filter1 is None
    assert filter2 is not None

    set_filter2(df.x > 4)
    assert filter1 is not None
    assert filter2 is not None

    set_filter2(None)
    assert filter1 is None
    assert filter2 is not None


def test_remove():
    set_filter1 = filter1 = None
    set_filter2 = filter2 = None
    set_multiple = None
    cross_filter_store: Optional[solara.hooks.dataframe.CrossFilterStore] = None

    @solara.component
    def FilterDummy1():
        nonlocal filter1, set_filter1
        filter1, set_filter1 = solara.use_cross_filter(id(df), "test1")
        return solara.Text("dummy1")

    @solara.component
    def FilterDummy2():
        nonlocal filter2, set_filter2
        filter2, set_filter2 = solara.use_cross_filter(id(df), "test1")
        return solara.Text("dummy2")

    @solara.component
    def Page():
        nonlocal cross_filter_store, set_multiple
        multiple, set_multiple = solara.use_state(True)
        cross_filter_store = solara.provide_cross_filter()
        with solara.VBox() as main:
            if multiple:
                FilterDummy1()
                FilterDummy2()
            else:
                FilterDummy1()
        return main

    rc, box = solara.render(Page(), handle_error=False)
    assert set_multiple is not None
    assert set_filter1 is not None
    assert set_filter2 is not None
    assert cross_filter_store is not None
    set_filter1(df.x > 1)
    filters = cross_filter_store.filters[id(df)]
    assert len(filters) == 1
    set_filter2(df.x < 4)
    assert len(filters) == 2
    set_multiple(False)
    assert len(filters) == 1
    set_multiple(True)
    # trigger a remove which did not add a filter
    set_multiple(False)
    assert len(filters) == 1



import pytest
import solara
import vaex
import polars as pl
from solara.components.datatable import DataTable, DataTableWidget

df_vaex = vaex.datasets.titanic()
df_pandas = df_vaex.to_pandas_df()
df_polars = pl.from_pandas(df_pandas)


@pytest.mark.parametrize("df", [df_vaex, df_pandas, df_polars])
def test_render(df):
    @solara.component
    def Test():
        return DataTable(df)

    widget, rc = solara.render_fixed(Test(), handle_error=False)
    assert isinstance(widget, DataTableWidget)
    assert len(widget.items) == 20



from pathlib import Path

import playwright
import playwright.sync_api

app_path = Path(__file__).parent / "testapp.py"


def test_error_in_render(page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    with extra_include_path(app_path.parent), solara_app("testapp:clickboom"):
        page_session.goto(solara_server.base_url)
        page_session.locator("text=Boom").click()
        page_session.locator("text=I crash on 1").wait_for()



from typing import List, cast

import numpy as np
import reacton.bqplot as bqplot
import reacton.ipyvuetify as v
import reacton.ipywidgets as w

import solara
from solara.components import ui_checkbox, ui_dropdown
from solara.hooks import use_cross_filter
from solara.lab.hooks.dataframe import use_df_column_names
from solara.lab.utils.dataframe import df_unique

cardheight = "100%"


@solara.component
def ExpressionEditor(df, value: str, label="Custom expression", on_value=None, placeholder="Enter an expression", prepend_icon="function"):
    """Editor for expression for Vaex, on_value will only be triggered for valid expressions"""

    def get_error(value):
        if value is None:
            return None
        try:
            df.validate_expression(value)
            return None
        except Exception as e:
            return str(e)

    value, set_value = solara.use_state(value)
    error: List = []
    error, set_error = solara.use_state(cast(List, []))
    set_error([get_error(value)])

    def on_value_local(value):
        error = get_error(value)
        set_value(value)
        set_error([error])

        if not error and on_value:
            on_value(df[value])

    return v.TextField(
        label=label,
        v_model=value,
        on_v_model=on_value_local,
        placeholder=placeholder,
        prepend_icon="mdi-filter",
        error_messages=error,
        success_messages="Looking good" if value is not None else None,
    )


@solara.component
def FilterCard(df):
    filter, set_filter = use_cross_filter(id(df), "filter-custom")

    with v.Card(elevation=2, height=cardheight) as main:
        with v.CardTitle(children=["Filter"]):
            pass
        with v.CardText():
            ExpressionEditor(df, "", on_value=set_filter)
    return main


@solara.component
def Table(df, n=5):
    output_c = w.Output()

    def output():
        import ipywidgets as widgets

        output_real: widgets.Output = solara.get_widget(output_c)
        # don't rely on display, does not work in solara yet
        with output_real.hold_sync():
            output_real.outputs = tuple()
            output_real.append_display_data(df)

    solara.use_side_effect(output)
    return output_c


@solara.component
def TableCard(df):
    filter, set_filter = use_cross_filter(id(df), "table")
    dff = df
    filtered = False
    if filter is not None:
        filtered = True
        dff = df[filter]
    if filtered:
        title = "Filtered"
    else:
        title = "Showing all"
    title = "Showing first 10 rows"
    progress = len(dff) / len(df) * 100
    with v.Card(elevation=2, height=cardheight) as main:
        with v.CardTitle(children=[title]):
            if filtered:
                v.ProgressLinear(value=progress)
        with v.CardText():
            Table(dff)
    return main


@solara.component
def HistogramCard(df, column=None, max_unique=100):
    filter, set_filter = use_cross_filter(id(df), "filter-histogram")
    dff = df  # filter(df)

    items = use_df_column_names(df)

    with v.Card(elevation=2, height=cardheight) as main:
        with v.CardTitle(children=["Histogram"]):
            pass
        with v.CardText():
            with v.Btn(v_on="x.on", icon=True, absolute=True, style_="right: 10px; top: 10px") as btn:
                v.Icon(children=["mdi-settings"])
            with v.Dialog(v_slots=[{"name": "activator", "variable": "x", "children": btn}], width="500"):
                with v.Sheet():
                    with v.Card(elevation=2):
                        with v.CardTitle(children=["Histogram input"]):
                            pass
                        with v.CardText():
                            column = items[0] if column not in items else column
                            column = ui_dropdown(value=column, label="x", options=items)
            if column:
                log = False
                import vaex

                dfg = dff.groupby(column, agg={"count": vaex.agg.count(selection=filter)}, sort=True)
                if len(dfg) > max_unique:
                    with v.Alert(
                        type="warning",
                        text=True,
                        prominent=True,
                        icon="mdi-alert",
                        children=[f"Too many unique values: {len(dfg)}, only showing first {max_unique}"],
                    ):
                        pass
                    dfg = dfg[:max_unique]
                if 1:
                    x = dfg[column].to_numpy()
                    y = dfg["count"].to_numpy()
                    if df[column].dtype == bool:
                        scale_x = bqplot.OrdinalScale()
                        x = np.where(x.astype("int8"), "true", "false")
                    elif dfg[column].dtype == str:
                        scale_x = bqplot.OrdinalScale(domain=x.tolist())
                    else:
                        scale_x = bqplot.OrdinalScale()
                    if log:
                        scale_y = bqplot.LogScale()
                    else:
                        scale_y = bqplot.LinearScale()

                    def on_selected(selected):
                        if selected is not None:
                            if df[column].dtype == bool:
                                value = [True, False][selected[0]]
                            else:
                                value = x[selected[0]]
                            expression = df[column] == value
                            set_filter(expression)
                        else:
                            set_filter(None)

                    lines = bqplot.Bars(
                        colors=["rgba(58, 130, 246, 0.5)"],
                        x=x.tolist(),
                        y=y.tolist(),
                        scales={"x": scale_x, "y": scale_y},
                        type="grouped",
                        selected_style={"fill": "rgba(58, 130, 246, 0.8)"},
                        on_selected=on_selected,
                        interactions={"click": "select"},
                    )
                    x_axis = bqplot.Axis(
                        scale=scale_x,
                        label=column,
                        grid_lines="none",
                        color="rgba(0,0,0,0)",
                        tick_style={
                            "fill": "rgba(0,0,0,0.8)",
                        },
                    )
                    y_axis = bqplot.Axis(
                        scale=scale_y,
                        orientation="vertical",
                        label="count",
                        color="rgba(0,0,0,0)",
                        tick_style={
                            "fill": "rgba(0,0,0,0.8)",
                        },
                    )
                    axes = [x_axis, y_axis]
                    bqplot.Figure(axes=axes, marks=[lines])
    return main


@solara.component
def ScatterCard(df, x=None, y=None, color=None):
    filter, set_filter = use_cross_filter(id(df), "filter-scatter")
    dff = df
    if filter:
        dff = df[filter]
    columns = df.get_column_names()
    max_points = 1000
    floats = [k for k in columns if df[k].dtype == float]
    #     flots = df.get_column_names()

    with v.Card(elevation=2, height=cardheight) as main:
        with v.CardTitle(children=["Scatter"]):
            pass
        with v.CardText():
            with v.Btn(v_on="x.on", icon=True, absolute=True, style_="right: 10px; top: 10px") as btn:
                v.Icon(children=["mdi-settings"])
            with v.Dialog(v_slots=[{"name": "activator", "variable": "x", "children": btn}], width="500"):
                with v.Sheet():
                    with v.Card(elevation=2):
                        with v.CardTitle(children=["Histogram input"]):
                            pass
                        with v.CardText():
                            xcol = x
                            if xcol is None:
                                xcol = xcol if len(floats) == 0 else floats[0]
                            xcol = ui_dropdown(value=xcol, label="x", options=floats)
                            ycol = y
                            if ycol is None:
                                ycol = ycol if len(floats) < 2 else floats[1]
                            ycol = ui_dropdown(value=ycol, label="y", options=floats)
                            ccol = color
                            if ccol is None:
                                ccol = ccol if len(floats) < 3 else floats[0]
                            ccol = ui_dropdown(value=ccol, label="Color", options=floats)
            if xcol and ycol:
                if len(dff) > max_points:
                    v.Alert(
                        type="warning", text=True, prominent=True, icon="mdi-alert", children=[f"Too many unique values, will only show first {max_points}"]
                    )
                    dff = dff[:max_points]

                x = dff[xcol].to_numpy()
                y = dff[ycol].to_numpy()
                if ccol is not None:
                    colord = dff[ccol].to_numpy().tolist()
                else:
                    colord = None

                if df[xcol].dtype == bool:
                    scale_x = bqplot.OrdinalScale()
                    x = np.where(x.astype("int8"), "true", "false")
                elif df[xcol].dtype == str:
                    scale_x = bqplot.OrdinalScale()
                else:
                    scale_x = bqplot.LinearScale()

                if df[ycol].dtype == bool:
                    scale_y = bqplot.OrdinalScale()
                    y = np.where(y.astype("int8"), "true", "false")
                elif df[ycol].dtype == str:
                    scale_y = bqplot.OrdinalScale()
                else:
                    scale_y = bqplot.LinearScale()

                def on_selected(selected):
                    if selected is not None and len(selected) > 0:
                        if df[xcol].dtype == bool:
                            value_x = [True, False][selected[0]]
                        else:
                            value_x = x[selected[0]]
                        if df[ycol].dtype == bool:
                            value_y = [True, False][selected[0]]
                        else:
                            value_y = y[selected[0]]
                        expression = (df[xcol] == value_x) & (df[ycol] == value_y)
                        set_filter(expression)
                    else:
                        set_filter(None)

                scale_color = bqplot.ColorScale()
                scatter = bqplot.Scatter(
                    x=x,
                    y=y,
                    scales={"x": scale_x, "y": scale_y, "color": scale_color},
                    color=colord,
                    selected_style={"fill": "#f55"},
                    on_selected=on_selected,
                    interactions={"click": "select"},
                )
                x_axis = bqplot.Axis(scale=scale_x, label=xcol)
                y_axis = bqplot.Axis(scale=scale_y, orientation="vertical", label=ycol)
                axes = [x_axis, y_axis]
                marks = [scatter]
                bqplot.Figure(axes=axes, marks=marks)
    return main


color_maps = [
    "Spectral",
    "RdYlGn",
    "RdBu",
    "PiYG",
    "PRGn",
    "RdYlBu",
    "BrBG",
    "RdGy",
    "PuOr",
    "Set2",
    "Accent",
    "Set1",
    "Set3",
    "Dark2",
    "Paired",
    "Pastel2",
    "Pastel1",
    "OrRd",
    "PuBu",
    "BuPu",
    "Oranges",
    "BuGn",
    "YlOrBr",
    "YlGn",
    "Reds",
    "RdPu",
    "Greens",
    "YlGnBu",
    "Purples",
    "GnBu",
    "Greys",
    "YlOrRd",
    "PuRd",
    "Blues",
    "PuBuGn",
    "viridis",
    "plasma",
    "inferno",
    "magma",
]

cardheight = "100%"


@solara.component
def HeatmapCard(df, x=None, y=None, debounce=True):
    limit_keys = "xmin xmax ymin ymax".split()
    limits, set_limits = solara.use_state(dict.fromkeys(limit_keys))
    contrast, set_contrast = solara.use_state([0.5, 99.5])
    # print("limits", limits)
    limits = limits.copy()
    filter, set_filter = use_cross_filter(id(df), "filter-heatmap")
    dff = df
    selection = filter
    # print("unfiltered", dff is df)

    items = df.get_column_names()
    floats = [k for k in items if df[k].dtype == float]

    with v.Card(elevation=2, height=cardheight) as main:
        with v.CardTitle(children=["Heatmap"]):
            pass
        with v.CardText():
            with v.Btn(v_on="x.on", icon=True, absolute=True, style_="right: 10px; top: 10px") as btn:
                v.Icon(children=["mdi-settings"])
            with v.Dialog(v_slots=[{"name": "activator", "variable": "x", "children": btn}], width="700"):
                with v.Sheet():
                    with v.Card(elevation=2):
                        with v.CardTitle(children=["Histogram input"]):
                            pass
                        with v.CardText():
                            xcol = x
                            if xcol is None:
                                xcol = xcol if len(floats) == 0 else floats[0]
                            xcol = ui_dropdown(value=xcol, label="x", options=floats)
                            ycol = y
                            if ycol is None:
                                ycol = ycol if len(floats) == 0 else floats[1]
                            ycol = ui_dropdown(value=ycol, label="y", options=floats)
                            scheme = ui_dropdown("Color map", value=color_maps[0], options=color_maps)
                            v.RangeSlider(v_model=contrast, on_v_model=set_contrast, label="Contrast", min=0, max=100)
                            crossfilter_visible = ui_checkbox(value=False, label="Cross filter visible")

            if xcol and ycol:
                import vaex.jupyter

                def updater(name):
                    def make():
                        # def shared():  # all render functions get a shared function per 'name'
                        def setter(value):
                            def state_updater(state):
                                return {**state, name: value}

                            set_limits(state_updater)

                        if debounce:
                            return vaex.jupyter.debounced(0.3)(setter)
                        else:
                            return setter
                        # return shared

                    return solara.use_memo(make, [name])

                update_xmin = updater("xmin")
                update_xmax = updater("xmax")
                update_ymin = updater("ymin")
                update_ymax = updater("ymax")

                # def update(name):
                #     def update_single(value):
                #         update_single_debouced(name, value)
                #     return update_single

                def minx(x):
                    values = [k.item() for k in dff[x].minmax()]
                    update_xmin(values[0])
                    update_xmax(values[1])

                solara.use_memo(lambda: minx(xcol), [xcol])

                # @solara.use_memo
                def miny(y):
                    values = [k.item() for k in dff[y].minmax()]
                    update_ymin(values[0])
                    update_ymax(values[1])

                solara.use_memo(lambda: miny(ycol), [ycol])

                if all(limits[k] is not None for k in limit_keys):
                    # print("limits used", limits)
                    xrange = limits["xmin"], limits["xmax"]
                    yrange = limits["ymin"], limits["ymax"]

                    def cross_filter():
                        if crossfilter_visible:
                            visible_filter = (df[xcol] >= xrange[0]) & (df[xcol] <= xrange[1]) & (df[ycol] >= yrange[0]) & (df[ycol] <= yrange[1])
                            set_filter(visible_filter)
                        else:
                            set_filter(None)

                    solara.use_memo(cross_filter, [crossfilter_visible, xrange, yrange])

                    # @solara.use_memo

                    def grid(xcol, ycol, limits):
                        vaex_limits = [xrange, yrange]
                        # print(xcol, ycol, vaex_limits)
                        # print("grid", len(dff))
                        return dff.count(binby=(xcol, ycol), limits=vaex_limits, shape=(512, 256), selection=selection)

                    values = grid(xcol, ycol, limits).astype("float32").T

                    vmin = np.percentile(values.ravel(), contrast[0]).item()
                    vmax = np.percentile(values.ravel(), contrast[1]).item()
                    if vmin == vmax:
                        vmax = values.min().item()
                        vmax = values.max().item()
                    # print(values.min(), values.max(), vmin, vmax, contrast)

                    scale_x = bqplot.LinearScale(allow_padding=False, on_min=update_xmin, on_max=update_xmax)  # min=0, max=1)
                    scale_y = bqplot.LinearScale(allow_padding=False, on_min=update_ymin, on_max=update_ymax)  # min=0, max=1)
                    scales = {"x": scale_x, "y": scale_y, "image": bqplot.ColorScale(min=vmin, max=vmax, scheme=scheme)}
                    from bqplot_image_gl import ImageGL

                    image = ImageGL.element(image=values, scales=scales, x=xrange, y=yrange)
                    panzoom = bqplot.PanZoom(scales={"x": [scales["x"]], "y": [scales["y"]]})

                    x_axis = bqplot.Axis(scale=scale_x, label=xcol)
                    y_axis = bqplot.Axis(scale=scale_y, orientation="vertical", label=ycol)
                    axes = [x_axis, y_axis]
                    marks = [image]
                    bqplot.Figure(axes=axes, marks=marks, interaction=panzoom)
    return main


@solara.component
def SummaryCard(df):
    filter, set_filter = use_cross_filter(id(df), "summary")
    dff = df
    filtered = False
    if filter is not None:
        filtered = True
        dff = df[filter]
    if filtered:
        title = "Filtered"
    else:
        title = "Showing all"
    progress = len(dff) / len(df) * 100
    with v.Card(elevation=2, height=cardheight) as main:
        with v.CardTitle(children=[title]):
            if filtered:
                v.ProgressLinear(value=progress)
        with v.CardText():
            icon = "mdi-filter"
            v.Icon(children=[icon], style_="opacity: 0.1" if not filtered else "")
            if filtered:
                summary = f"{len(dff):,} / {len(df):,}"
            else:
                summary = f"{len(dff):,}"
            v.Html(tag="h3", children=[summary], style_="display: inline")
    return main


@solara.component
def DropdownCard(df, column=None):
    max_unique = 100
    filter, set_filter = use_cross_filter(id(df), "filter-dropdown")
    columns = use_df_column_names(df)
    column, set_column = solara.use_state(columns[0] if column is None else column)
    uniques = df_unique(df, column, limit=max_unique + 1)
    value, set_value = solara.use_state(None)
    # to avoid confusing vuetify about selecting 'None' and nothing
    magic_value_missing = "__missing_value__"

    def set_value_and_filter(value):
        set_value(value)
        # print(value)
        if value is None:
            set_filter(None)
        else:
            value = value["value"]
            if value == magic_value_missing:
                set_filter(str(df[column].ismissing()))
            else:
                filter = df[column] == value
                # print(filter)
                set_filter(filter)

    with v.Card(elevation=2, height=cardheight) as main:
        with v.CardTitle(children=["Filter out single value"]):
            pass
        with v.CardText():
            with v.Btn(v_on="x.on", icon=True, absolute=True, style_="right: 10px; top: 10px") as btn:
                v.Icon(children=["mdi-settings"])
            with v.Dialog(v_slots=[{"name": "activator", "variable": "x", "children": btn}], width="500"):
                with v.Sheet():
                    with v.Container(pa_4=True, ma_0=True):
                        with v.Row():
                            with v.Col():
                                v.Select(v_model=column, items=columns, on_v_model=set_column, label="Choose column")
            # we use objects to we can distinguish between selecting nothing or None
            items = [{"value": magic_value_missing if k is None else k, "text": str(k)} for k in uniques]
            v.Select(v_model=value, items=items, on_v_model=set_value_and_filter, label=f"Choose {column} value", clearable=True, return_object=True)
            if len(uniques) > max_unique:
                v.Alert(type="warning", text=True, prominent=True, icon="mdi-alert", children=[f"Too many unique values, will only show first {max_unique}"])

    return main



"""# generate_routes_directory

"""
import solara
import solara.autorouting
from solara.website.utils import apidoc

from . import NoPage

title = "generate_routes_directory"
Page = NoPage
__doc__ += apidoc(solara.autorouting.generate_routes_directory)  # type: ignore



"""# Countdown timer.

This example shows how to use [use_thread](/api/use_thread) to create a countdown timer.

The UI code demonstrates a lot of conditional rendering.

"""

import time

import solara
from solara.alias import rv


@solara.component
def Page():
    seconds, set_seconds = solara.use_state(5)
    running, set_running = solara.use_state(False)
    duration, set_duration = solara.use_state(seconds)

    def on_duration(duration):
        try:
            set_duration(int(duration))
            set_seconds(int(duration))
        except ValueError:
            pass

    def run_timer():
        if running:
            time_start = time.time()
            next_tick = time_start + 1
            i = seconds
            while i:
                # instead of sleeping for 1 second, we sleep until the next second
                # this takes into account overhead and makes the timer more accurate
                time.sleep(max(0, next_tick - time.time()))
                i -= 1
                set_seconds(i)
                next_tick += 1
            set_running(False)

    solara.use_thread(run_timer, dependencies=[duration, running])

    with solara.VBox() as main:
        if not running:
            if duration < 1:
                solara.Error("Duration must be at least 1 second")
            else:
                solara.Markdown(f"# Timer set to {seconds} seconds")
        else:
            if seconds:
                solara.Markdown(f"# {seconds} seconds left")
            else:
                solara.solara.Markdown("# Time's up!")
        rv.TextField(type="number", v_model=duration, on_v_model=on_duration, disabled=running)
        with solara.HBox():
            if running:
                solara.Button("Stop", on_click=lambda: set_running(False), icon_name="mdi-stop")
            else:
                if duration != seconds:
                    solara.Button("Reset", on_click=lambda: set_seconds(duration), icon_name="mdi-restart")
                else:
                    solara.Button("Start", on_click=lambda: set_running(True), icon_name="mdi-play", disabled=seconds < 1)
    return main



"""# Route

"""
import solara
from solara.website.utils import apidoc

routes = [
    solara.Route(path="/"),
    solara.Route(path="kiwi"),
    solara.Route(path="banana"),
    solara.Route(path="apple"),
]


@solara.component
def Page():
    route_current, routes = solara.use_route()
    with solara.VBox() as main:
        # solara.Warning("Note the address bar in the browser. It should change to the path of the link.")
        solara.Markdown("*Click on one of the links below to change the route and see the url in your browser change, and match the route.*")
        with solara.VBox():
            for route in routes:
                with solara.Link(route):
                    current = route_current is route
                    if current:
                        solara.Success(f"You are at solara.Route(path={route.path!r})")
                    else:
                        solara.Info(f"Go to solara.Route(path={route.path!r})")
    return main


__doc__ += apidoc(solara.Route, full=True)  # type: ignore



import asyncio
import time

import ipyvuetify as v
import pytest
from reacton import ipywidgets as w

import solara.tasks
from solara.server import kernel, kernel_context
from solara.tasks import TaskState, use_task
from solara.toestand import Computed


@solara.tasks.task
def something(count: int, delay: float = 0.1):
    time.sleep(delay)
    return "42" * count


@solara.component
def ComputeButton(count, delay: float = 0.1, on_render=lambda: None):
    solara.Button("Run", on_click=lambda: something(count, delay))
    on_render()
    # print(something.result.value)
    if something.result.value:
        if something.pending:
            solara.Info("running")
        elif something.finished:
            solara.Info("Done: " + str(something.value))
        elif something.error:
            solara.Info("Error: " + str(something.exception))
        elif something.cancelled:
            solara.Info("Cancelled")
        elif something.not_called:
            solara.Info("Not called yet")
        else:
            raise RuntimeError("should not happen")


@solara.component
def Page():
    ComputeButton(2)
    ComputeButton(3)


cancel_square = False


@solara.tasks.task
def square(value: float):
    if cancel_square:
        square.cancel()
    return value**2


@solara.component
def SquareButton(value, on_render=lambda: None):
    solara.Button("Run", on_click=lambda: square(value))
    on_render()
    if square.result.value:
        if square.pending:
            solara.Info("running")
        elif square.finished:
            solara.Info("Done: " + str(square.value))
        elif square.error:
            solara.Info("Error: " + str(square.error))
        elif square.cancelled:
            solara.Info("Cancelled")
        elif square.not_called:
            solara.Info("Not called yet")
        else:
            raise RuntimeError("should not happen")


def test_task_basic():
    results = []

    def collect():
        results.append((square._result.value._state, square.latest))

    box, rc = solara.render(SquareButton(3, on_render=collect), handle_error=False)
    button = rc.find(v.Btn, children=["Run"]).widget
    button.click()
    assert square._last_finished_event  # type: ignore
    square._last_finished_event.wait()  # type: ignore
    assert results == [
        (TaskState.NOTCALLED, None),
        (TaskState.STARTING, None),
        (TaskState.RUNNING, None),
        (TaskState.FINISHED, 9),
    ]
    results.clear()
    rc.render(SquareButton(2, on_render=collect))
    button = rc.find(v.Btn, children=["Run"]).widget
    button.click()
    square._last_finished_event.wait()  # type: ignore
    assert results == [
        # extra finished due to the rc.render call
        (TaskState.FINISHED, 9),
        (TaskState.STARTING, 9),
        (TaskState.RUNNING, 9),
        (TaskState.FINISHED, 4),
    ]


# async version

cancel_square_async = False


@solara.tasks.task
async def square_async(value: float):
    if cancel_square_async:
        square_async.cancel()
    return value**2


@solara.component
def SquareButtonAsync(value, on_render=lambda: None):
    solara.Button("Run", on_click=lambda: square_async(value))
    on_render()
    if square_async.result.value:
        if square_async.pending:
            solara.Info("running")
        elif square_async.finished:
            solara.Info("Done: " + str(square_async.value))
        elif square_async.error:
            solara.Info("Error: " + str(square_async.exception))
        elif square_async.cancelled:
            solara.Info("Cancelled")
        elif square_async.not_called:
            solara.Info("Not called yet")
        else:
            raise RuntimeError("should not happen")


@pytest.mark.asyncio
@pytest.mark.parametrize("run_in_thread", [True, False])
async def test_task_basic_async(run_in_thread):
    results = []
    assert square_async._instance.value.run_in_thread  # type: ignore
    square_async._instance.value.run_in_thread = run_in_thread  # type: ignore

    def collect():
        results.append((square_async._result.value._state, square_async.latest))

    box, rc = solara.render(SquareButtonAsync(3, on_render=collect), handle_error=False)
    button = rc.find(v.Btn, children=["Run"]).widget
    button.click()
    assert square_async.current_future  # type: ignore
    await square_async.current_future  # type: ignore
    assert results == [
        (TaskState.NOTCALLED, None),
        (TaskState.STARTING, None),
        (TaskState.RUNNING, None),
        (TaskState.FINISHED, 9),
    ]
    results.clear()
    rc.render(SquareButtonAsync(2, on_render=collect))
    button = rc.find(v.Btn, children=["Run"]).widget
    button.click()
    await square_async.current_future  # type: ignore
    assert results == [
        # extra finished due to the rc.render call
        (TaskState.FINISHED, 9),
        (TaskState.STARTING, 9),
        (TaskState.RUNNING, 9),
        (TaskState.FINISHED, 4),
    ]
    square_async._instance.value.run_in_thread = True  # type: ignore


def test_task_two():
    results2 = []
    results3 = []
    # ugly reset
    square._last_value = None

    def collect2():
        results2.append((square._result.value._state, square.latest))

    def collect3():
        results3.append((square._result.value._state, square.latest))

    @solara.component
    def Test():
        SquareButton(2, on_render=collect2)
        SquareButton(3, on_render=collect3)

    box, rc = solara.render(Test(), handle_error=False)
    button = rc.find(v.Btn, children=["Run"])[0].widget
    button.click()
    assert square._last_finished_event  # type: ignore
    square._last_finished_event.wait()  # type: ignore
    assert (
        results2
        == results3
        == [
            (TaskState.NOTCALLED, None),
            (TaskState.STARTING, None),
            (TaskState.RUNNING, None),
            (TaskState.FINISHED, 4),
        ]
    )
    assert len(rc.find(children=["Done: 4"])) == 2

    # now we press the second button
    results2.clear()
    results3.clear()
    button = rc.find(v.Btn, children=["Run"])[1].widget
    button.click()
    assert square._last_finished_event  # type: ignore
    square._last_finished_event.wait()  # type: ignore
    assert (
        results2
        == results3
        == [
            # not a finished event, because we don't render from the start
            (TaskState.STARTING, 4),
            (TaskState.RUNNING, 4),
            (TaskState.FINISHED, 9),
        ]
    )
    assert len(rc.find(children=["Done: 9"])) == 2


def test_task_cancel_retry():
    global cancel_square
    results = []

    # ugly reset
    square._last_value = None

    def collect():
        results.append((square._result.value._state, square.value))

    box, rc = solara.render(SquareButton(5, on_render=collect), handle_error=False)
    button = rc.find(v.Btn, children=["Run"]).widget
    cancel_square = True
    try:
        button.click()
        assert square._last_finished_event  # type: ignore
        square._last_finished_event.wait()  # type: ignore
        assert results == [
            (TaskState.NOTCALLED, None),
            (TaskState.STARTING, None),
            (TaskState.RUNNING, None),
            (TaskState.CANCELLED, None),
        ]
    finally:
        cancel_square = False
    results.clear()
    square.retry()
    square._last_finished_event.wait()  # type: ignore
    assert results == [
        (TaskState.STARTING, None),
        (TaskState.RUNNING, None),
        (TaskState.FINISHED, 5**2),
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("run_in_thread", [True, False])
async def test_task_async_cancel_retry(run_in_thread):
    global cancel_square_async
    results = []

    assert square_async._instance.value.run_in_thread  # type: ignore
    square_async._instance.value.run_in_thread = run_in_thread  # type: ignore

    # ugly reset
    square_async._last_value = None

    def collect():
        results.append((square_async._result.value._state, square_async.value))

    box, rc = solara.render(SquareButtonAsync(5, on_render=collect), handle_error=False)
    button = rc.find(v.Btn, children=["Run"]).widget
    cancel_square_async = True
    try:
        button.click()
        assert square_async.current_future  # type: ignore
        try:
            await square_async.current_future  # type: ignore
        except asyncio.CancelledError:
            pass

        assert results == [
            (TaskState.NOTCALLED, None),
            (TaskState.STARTING, None),
            (TaskState.RUNNING, None),
            (TaskState.CANCELLED, None),
        ]
    finally:
        cancel_square_async = False
    results.clear()
    square_async.retry()
    await square_async.current_future  # type: ignore
    assert results == [
        (TaskState.STARTING, None),
        (TaskState.RUNNING, None),
        (TaskState.FINISHED, 5**2),
    ]

    square_async._instance.value.run_in_thread = True  # type: ignore


def test_task_scopes(no_kernel_context):
    results1 = []
    results2 = []

    def collect1():
        results1.append((something._result.value._state, something.value))

    def collect2():
        results2.append((something._result.value._state, something.value))

    kernel1 = kernel.Kernel()
    kernel2 = kernel.Kernel()
    assert kernel_context.current_context[kernel_context.get_current_thread_key()] is None

    context1 = kernel_context.VirtualKernelContext(id="toestand-1", kernel=kernel1, session_id="session-1")
    context2 = kernel_context.VirtualKernelContext(id="toestand-2", kernel=kernel2, session_id="session-2")

    with context1:
        box1, rc1 = solara.render(ComputeButton(5, on_render=collect1), handle_error=False)
        button1 = rc1.find(v.Btn, children=["Run"]).widget

    with context2:
        box2, rc2 = solara.render(ComputeButton(5, on_render=collect2), handle_error=False)
        button2 = rc2.find(v.Btn, children=["Run"]).widget

    with context1:
        button1.click()
        finished_event1 = something._last_finished_event  # type: ignore
        assert finished_event1

    with context2:
        assert something._last_finished_event is not finished_event1  # type: ignore
        assert something._last_finished_event is None  # type: ignore

    finished_event1.wait()
    assert results1 == [
        (TaskState.NOTCALLED, None),
        (TaskState.STARTING, None),
        (TaskState.RUNNING, None),
        (TaskState.FINISHED, "4242424242"),
    ]
    # results1.clear()
    assert results2 == [(TaskState.NOTCALLED, None)]

    with context2:
        button2.click()
        finished_event2 = something._last_finished_event  # type: ignore
        assert finished_event2
    finished_event2.wait()
    assert results2 == [
        (TaskState.NOTCALLED, None),
        (TaskState.STARTING, None),
        (TaskState.RUNNING, None),
        (TaskState.FINISHED, "4242424242"),
    ]


def test_task_and_computed(no_kernel_context):
    called = 0

    @Computed
    def square_minus_one():
        nonlocal called
        called += 1
        return square.latest - 1

    kernel1 = kernel.Kernel()
    kernel2 = kernel.Kernel()
    assert kernel_context.current_context[kernel_context.get_current_thread_key()] is None

    context1 = kernel_context.VirtualKernelContext(id="t1", kernel=kernel1, session_id="session-1")
    context2 = kernel_context.VirtualKernelContext(id="t2", kernel=kernel2, session_id="session-2")

    with context1:
        r1 = square._result
        assert len(square._result._storage.listeners2["t1"]) == 0
        square(5)
        assert square._last_finished_event  # type: ignore
        square._last_finished_event.wait()  # type: ignore
        # accessing will add it to the listeners
        assert len(square._result._storage.listeners2["t1"]) == 0
        assert square_minus_one.value == 24
        assert called == 1
        assert len(square._result._storage.listeners2["t1"]) == 1
        # assert square_minus_one._auto_subscriber.value.reactive_used == {square.value}

    with context2:
        r2 = square._result
        assert len(square._result._storage.listeners2["t2"]) == 0
        square(6)
        assert square._last_finished_event  # type: ignore
        square._last_finished_event.wait()  # type: ignore
        assert len(square._result._storage.listeners2["t2"]) == 0
        assert square_minus_one.value == 35
        assert called == 2
        assert len(square._result._storage.listeners2["t2"]) == 1
        # square_minus_one._auto_subscriber.value.reactive_used == {square.value}

    with context1:
        assert r1 is square._result
        # assert len(square.result._storage.listeners2["t1"]) == 1
        square._last_finished_event = None  # type: ignore
        square_minus_one._auto_subscriber.value.reactive_used == {square.value}
        assert square_minus_one.value == 24
        assert called == 2
        square(7)
        square_minus_one._auto_subscriber.value.reactive_used == {square.value}
        assert square._last_finished_event  # type: ignore
        square._last_finished_event.wait()  # type: ignore
        assert square_minus_one.value == 48
        assert called == 3

    with context2:
        assert r2 is square._result
        assert square_minus_one.value == 35
        square(8)
        assert square._last_finished_event  # type: ignore
        square._last_finished_event.wait()  # type: ignore
        assert square_minus_one.value == 63
        assert called == 4


# copied from hooks_test.py


def test_use_task_intrusive_cancel():
    task = None
    last_value = 0
    seconds = 4.0

    def retry():
        pass

    @solara.component
    def Test():
        nonlocal task
        nonlocal last_value
        retry_counter, set_retry_counter = solara.use_state(0)

        nonlocal retry

        def local_retry():
            set_retry_counter(lambda x: x + 1)

        retry = local_retry

        def work():
            nonlocal last_value
            nonlocal task
            assert task is not None
            for i in range(100):
                last_value = i
                # if not cancelled, might take 4 seconds
                time.sleep(seconds / 100)
                if not task.is_current():
                    return

            return 2**42

        task = use_task(work, dependencies=[retry_counter])
        assert task is not None
        return w.Label(value="test")

    solara.render_fixed(Test(), handle_error=False)
    assert task is not None
    # result.cancel()
    # while result._state in [TaskState.STARTING, TaskState.RUNNING]:
    #     time.sleep(0.1)
    # assert result._state == TaskState.CANCELLED
    # assert last_value != 99

    # also test retry
    while task._state != TaskState.RUNNING:
        time.sleep(0.05)
    assert last_value != 99
    seconds = 0.1
    retry()
    # wait till it stops running
    while task._state == TaskState.RUNNING:
        time.sleep(0.05)
    # wait till it exits these states
    while task._state in [TaskState.STARTING, TaskState.WAITING, TaskState.RUNNING]:
        time.sleep(0.1)
    assert task._state == TaskState.FINISHED
    assert last_value == 99


@pytest.mark.asyncio
@pytest.mark.parametrize("prefer_threaded", [True, False])
async def test_use_task_async(prefer_threaded):
    task = None
    last_value = 0
    seconds = 4.0

    def retry():
        pass

    @solara.component
    def Test():
        nonlocal task
        nonlocal last_value
        retry_counter, set_retry_counter = solara.use_state(0)
        nonlocal retry

        def local_retry():
            set_retry_counter(lambda x: x + 1)

        retry = local_retry

        async def work():
            nonlocal last_value
            assert task is not None
            for i in range(100):
                last_value = i
                # if not cancelled, might take 4 seconds
                await asyncio.sleep(seconds / 100)
                if not task.is_current():
                    return
            return 2**42

        task = use_task(work, dependencies=[retry_counter], prefer_threaded=prefer_threaded)
        return w.Label(value="test")

    solara.render_fixed(Test(), handle_error=False)
    assert task is not None
    # we do not support cancel anymore in use_task
    # result.cancel()
    # the current implementation if cancel is direct, we so we not need the code below
    # n = 0
    # while result.state in [TaskState.NOTCALLED, TaskState.STARTING, TaskState.RUNNING]:
    #     await asyncio.sleep(0.1)
    #     n += 1
    #     if n == 100:
    #         raise TimeoutError("took too long, state = " + str(result.state))
    # assert result._state == TaskState.CANCELLED
    # assert last_value != 99

    # also test retry
    seconds = 0.1
    retry()
    n = 0
    while task._state == TaskState.CANCELLED:
        await asyncio.sleep(0.1)
        n += 1
        if n == 100:
            raise TimeoutError("took too long, state = " + str(task._state))
    n = 0
    while task._state in [TaskState.STARTING, TaskState.RUNNING]:
        await asyncio.sleep(0.1)
        n += 1
        if n == 100:
            raise TimeoutError("took too long, state = " + str(task._state))
    assert task._state == TaskState.FINISHED
    assert last_value == 99



from typing import Any, Callable

import ipyvuetify
import traitlets

import solara


class EchartsWidget(ipyvuetify.VuetifyTemplate):
    template_file = (__file__, "echarts.vue")

    attributes = traitlets.Dict(default_value=None, allow_none=True).tag(sync=True)

    maps = traitlets.Any({}).tag(sync=True)
    option = traitlets.Any({}).tag(sync=True)
    on_click = traitlets.Callable(None, allow_none=True)
    on_mouseover = traitlets.Callable(None, allow_none=True)
    # for performance, so we don't get events from the frontend we don't care about
    on_mouseover_enabled = traitlets.Bool(False).tag(sync=True)
    on_mouseout = traitlets.Callable(None, allow_none=True)
    # same, for performance
    on_mouseout_enabled = traitlets.Bool(False).tag(sync=True)

    def vue_on_click(self, data):
        if self.on_click:
            self.on_click(data)

    def vue_on_mouseover(self, data):
        if self.on_mouseover:
            self.on_mouseover(data)

    def vue_on_mouseout(self, data):
        if self.on_mouseout:
            self.on_mouseout(data)


@solara.component
def FigureEcharts(
    option: dict = {},
    on_click: Callable[[Any], Any] = None,
    on_mouseover: Callable[[Any], Any] = None,
    on_mouseout: Callable[[Any], Any] = None,
    maps: dict = {},
    attributes={"style": "height: 400px"},
):
    """Create a Echarts figure.

    See [The Echarts website for examples](https://echarts.apache.org/)

    Note that we do not support a Python API to create the figure data.

    A library such as Pyecharts can help you with that, otherwise you can provide
    the data simply as data similarly as on the Echarts example webpage.

    # Arguments

    * option: dict, the option for the figure, see the echart documentation.
    * on_click: Callable, a function that will be called when the user clicks on the figure.
    * on_mouseover: Callable, a function that will be called when the user moves the mouse over a certain component.
    * on_mouseout: Callable, a function that will be called when the user moves the mouse out of a certain component.
    * maps: dict, a dictionary of maps to be used in the figure.
    * attributes: dict, a dictionary of attributes to be passed to the container (like style, class).


    """
    return EchartsWidget.element(
        option=option,
        on_click=on_click,
        on_mouseover=on_mouseover,
        on_mouseover_enabled=on_mouseover is not None,
        maps=maps,
        on_mouseout=on_mouseout,
        on_mouseout_enabled=on_mouseout is not None,
        attributes=attributes,
    )



from unittest.mock import MagicMock

import ipyvuetify as vw

import solara


def test_chatinput():
    send_callback = MagicMock()
    el = solara.lab.ChatInput(send_callback=send_callback)
    box, rc = solara.render(el, handle_error=False)
    input = rc.find(vw.TextField)
    button = rc.find(vw.Btn)

    input.widget.v_model = "hello"
    assert send_callback.call_count == 0
    input.widget.fire_event("keyup.enter")
    assert send_callback.call_count == 1
    assert send_callback.call_args[0][0] == "hello"
    assert input.widget.v_model == ""
    input.widget.v_model = "hello"
    assert send_callback.call_count == 1
    button.widget.fire_event("click")
    assert send_callback.call_count == 2
    assert send_callback.call_args[0][0] == "hello"
    assert input.widget.v_model == ""



import ipywidgets as widgets
import playwright.sync_api
import pytest
from IPython.display import display

from .conftest import SERVERS

widgets_version_major = int(widgets.__version__.split(".")[0])
widgets_postfix = f"-ipywidgets-{widgets_version_major}"


def test_widget_latex_solara(solara_test, page_session: playwright.sync_api.Page, assert_solara_snapshot, request):
    if request.node.callspec.params["solara_server"] != SERVERS[0]:
        pytest.skip("No need to run this test for all servers.")
    # we tried a FloatSlider before, but that seems to give unstable images (handle moves around randomly)
    label = widgets.Label(value=r"$E \sim mc^2$")
    label.add_class("test-class-latex")
    container = widgets.VBox([label], layout={"width": "200px", "height": "100px"})
    display(container)

    page_session.locator(".test-class-latex >> .mrel").wait_for()
    page_session.evaluate("document.fonts.ready")
    page_session.wait_for_timeout(1000)
    assert_solara_snapshot(page_session.locator(".test-class-latex").screenshot(), postfix=widgets_postfix)
    label.value = r"$\alpha$"
    label.add_class("test-changed-class-latex")
    page_session.locator(".test-changed-class-latex >> .mathnormal").wait_for()
    page_session.evaluate("document.fonts.ready")
    page_session.wait_for_timeout(1000)
    assert_solara_snapshot(page_session.locator(".test-changed-class-latex").screenshot(), postfix=widgets_postfix + "-changed")
